@@ -211,6 +211,7 @@ from keystoneclient import exceptions
 from keystoneclient import session
 from oslo_config import cfg
 from oslo_serialization import jsonutils
+import pkg_resources
 import six
 import webob.dec
 
@@ -269,6 +270,8 @@ _OPTS = [
                help='A PEM encoded Certificate Authority to use when '
                     'verifying HTTPs connections. Defaults to system CAs.'),
     cfg.BoolOpt('insecure', default=False, help='Verify HTTPS connections.'),
+    cfg.StrOpt('region_name', default=None,
+               help='The region in which the identity server can be found.'),
     cfg.StrOpt('signing_dir',
                help='Directory used to cache files related to PKI tokens.'),
     cfg.ListOpt('memcached_servers',
@@ -412,6 +415,10 @@ def _conf_values_type_convert(conf):
     return opts
 
 
+def _get_project_version(project):
+    return pkg_resources.get_distribution(project).version
+
+
 class _BaseAuthProtocol(object):
     """A base class for AuthProtocol token checking implementations.
 
@@ -419,11 +426,17 @@ class _BaseAuthProtocol(object):
     :param logging.Logger log: The logging object to use for output. By default
                                it will use a logger in the
                                keystonemiddleware.auth_token namespace.
+    :param str enforce_token_bind: The style of token binding enforcement to
+                                   perform.
     """
 
-    def __init__(self, app, log=_LOG):
+    def __init__(self,
+                 app,
+                 log=_LOG,
+                 enforce_token_bind=_BIND_MODE.PERMISSIVE):
         self.log = log
         self._app = app
+        self._enforce_token_bind = enforce_token_bind
 
     @webob.dec.wsgify(RequestClass=_request._AuthTokenRequest)
     def __call__(self, req):
@@ -446,6 +459,76 @@ class _BaseAuthProtocol(object):
 
         By default this method does not return a value.
         """
+        request.remove_auth_headers()
+
+        user_auth_ref = None
+        serv_auth_ref = None
+
+        if request.user_token:
+            self.log.debug('Authenticating user token')
+            try:
+                data, user_auth_ref = self._do_fetch_token(request.user_token)
+                self._validate_token(user_auth_ref)
+                self._confirm_token_bind(user_auth_ref, request)
+            except exc.InvalidToken:
+                self.log.info(_LI('Invalid user token'))
+                request.user_token_valid = False
+            else:
+                request.user_token_valid = True
+                request.environ['keystone.token_info'] = data
+
+        if request.service_token:
+            self.log.debug('Authenticating service token')
+            try:
+                _, serv_auth_ref = self._do_fetch_token(request.service_token)
+                self._validate_token(serv_auth_ref)
+                self._confirm_token_bind(serv_auth_ref, request)
+            except exc.InvalidToken:
+                self.log.info(_LI('Invalid service token'))
+                request.service_token_valid = False
+            else:
+                request.service_token_valid = True
+
+        p = _user_plugin.UserAuthPlugin(user_auth_ref, serv_auth_ref)
+        request.environ['keystone.token_auth'] = p
+
+    def _validate_token(self, auth_ref):
+        """Perform the validation steps on the token.
+
+        :param auth_ref: The token data
+        :type auth_ref: keystoneclient.access.AccessInfo
+
+        :raises exc.InvalidToken: if token is rejected
+        """
+        # 0 seconds of validity means is it valid right now.
+        if auth_ref.will_expire_soon(stale_duration=0):
+            raise exc.InvalidToken(_('Token authorization failed'))
+
+    def _do_fetch_token(self, token):
+        """Helper method to fetch a token and convert it into an AccessInfo"""
+        data = self._fetch_token(token)
+
+        try:
+            return data, access.AccessInfo.factory(body=data, auth_token=token)
+        except Exception:
+            self.log.warning(_LW('Invalid token contents.'), exc_info=True)
+            raise exc.InvalidToken(_('Token authorization failed'))
+
+    def _fetch_token(self, token):
+        """Fetch the token data based on the value in the header.
+
+        Retrieve the data associated with the token value that was in the
+        header. This can be from PKI, contacting the identity server or
+        whatever is required.
+
+        :param str token: The token present in the request header.
+
+        :raises exc.InvalidToken: if token is invalid.
+
+        :returns: The token data
+        :rtype: dict
+        """
+        raise NotImplemented()
 
     def process_response(self, response):
         """Do whatever you'd like to the response.
@@ -454,6 +537,77 @@ class _BaseAuthProtocol(object):
         """
         return response
 
+    def _invalid_user_token(self, msg=False):
+        # NOTE(jamielennox): use False as the default so that None is valid
+        if msg is False:
+            msg = _('Token authorization failed')
+
+        raise exc.InvalidToken(msg)
+
+    def _confirm_token_bind(self, auth_ref, req):
+        if self._enforce_token_bind == _BIND_MODE.DISABLED:
+            return
+
+        try:
+            if auth_ref.version == 'v2.0':
+                bind = auth_ref['token']['bind']
+            elif auth_ref.version == 'v3':
+                bind = auth_ref['bind']
+            else:
+                self._invalid_user_token()
+        except KeyError:
+            bind = {}
+
+        # permissive and strict modes don't require there to be a bind
+        permissive = self._enforce_token_bind in (_BIND_MODE.PERMISSIVE,
+                                                  _BIND_MODE.STRICT)
+
+        if not bind:
+            if permissive:
+                # no bind provided and none required
+                return
+            else:
+                self.log.info(_LI('No bind information present in token.'))
+                self._invalid_user_token()
+
+        # get the named mode if bind_mode is not one of the predefined
+        if permissive or self._enforce_token_bind == _BIND_MODE.REQUIRED:
+            name = None
+        else:
+            name = self._enforce_token_bind
+
+        if name and name not in bind:
+            self.log.info(_LI('Named bind mode %s not in bind information'),
+                          name)
+            self._invalid_user_token()
+
+        for bind_type, identifier in six.iteritems(bind):
+            if bind_type == _BIND_MODE.KERBEROS:
+                if req.auth_type != 'negotiate':
+                    self.log.info(_LI('Kerberos credentials required and '
+                                      'not present.'))
+                    self._invalid_user_token()
+
+                if req.remote_user != identifier:
+                    self.log.info(_LI('Kerberos credentials do not match '
+                                      'those in bind.'))
+                    self._invalid_user_token()
+
+                self.log.debug('Kerberos bind authentication successful.')
+
+            elif self._enforce_token_bind == _BIND_MODE.PERMISSIVE:
+                self.log.debug('Ignoring Unknown bind for permissive mode: '
+                               '%(bind_type)s: %(identifier)s.',
+                               {'bind_type': bind_type,
+                                'identifier': identifier})
+
+            else:
+                self.log.info(
+                    _LI('Couldn`t verify unknown bind: %(bind_type)s: '
+                        '%(identifier)s.'),
+                    {'bind_type': bind_type, 'identifier': identifier})
+                self._invalid_user_token()
+
 
 class AuthProtocol(_BaseAuthProtocol):
     """Middleware that handles authenticating client calls."""
@@ -464,13 +618,38 @@ class AuthProtocol(_BaseAuthProtocol):
     def __init__(self, app, conf):
         log = logging.getLogger(conf.get('log_name', __name__))
         log.info(_LI('Starting Keystone auth_token middleware'))
-        super(AuthProtocol, self).__init__(app, log=log)
 
         # NOTE(wanghong): If options are set in paste file, all the option
         # values passed into conf are string type. So, we should convert the
         # conf value into correct type.
         self._conf = _conf_values_type_convert(conf)
 
+        # NOTE(sileht): If we don't want to use oslo.config global object
+        # we can set the paste "oslo_config_project" and the middleware
+        # will load the configuration with a local oslo.config object.
+        self._local_oslo_config = None
+        if 'oslo_config_project' in conf:
+            if 'oslo_config_file' in conf:
+                default_config_files = [conf['oslo_config_file']]
+            else:
+                default_config_files = None
+
+            self._local_oslo_config = cfg.ConfigOpts()
+            self._local_oslo_config(
+                {}, project=conf['oslo_config_project'],
+                default_config_files=default_config_files,
+                validate_default_values=True)
+
+            self._local_oslo_config.register_opts(
+                _OPTS, group=_base.AUTHTOKEN_GROUP)
+            auth.register_conf_options(self._local_oslo_config,
+                                       group=_base.AUTHTOKEN_GROUP)
+
+        super(AuthProtocol, self).__init__(
+            app,
+            log=log,
+            enforce_token_bind=self._conf_get('enforce_token_bind'))
+
         # delay_auth_decision means we still allow unauthenticated requests
         # through and we let the downstream service make the final decision
         self._delay_auth_decision = self._conf_get('delay_auth_decision')
@@ -512,6 +691,8 @@ class AuthProtocol(_BaseAuthProtocol):
         # try config from paste-deploy first
         if name in self._conf:
             return self._conf[name]
+        elif self._local_oslo_config:
+            return self._local_oslo_config[group][name]
         else:
             return CONF[group][name]
 
@@ -525,57 +706,39 @@ class AuthProtocol(_BaseAuthProtocol):
         depending on configuration.
         """
         self._token_cache.initialize(request.environ)
-        request.remove_auth_headers()
 
-        try:
-            user_auth_ref = None
-            serv_auth_ref = None
+        resp = super(AuthProtocol, self).process_request(request)
+        if resp:
+            return resp
 
-            try:
-                self.log.debug('Authenticating user token')
-                user_token_info = self._get_user_token_from_request(request)
-                user_auth_ref, user_token_info = self._validate_token(
-                    user_token_info, request)
-                request.environ['keystone.token_info'] = user_token_info
-                request.set_user_headers(user_auth_ref,
-                                         self._include_service_catalog)
-            except exc.InvalidToken:
-                if self._delay_auth_decision:
-                    self.log.info(
-                        _LI('Invalid user token - deferring reject '
-                            'downstream'))
-                    request.headers['X-Identity-Status'] = 'Invalid'
-                else:
-                    self.log.info(
-                        _LI('Invalid user token - rejecting request'))
-                    self._reject_request()
+        if not request.user_token:
+            # if no user token is present then that's an invalid request
+            request.user_token_valid = False
 
-            try:
-                self.log.debug('Authenticating service token')
-                if request.service_token is not None:
-                    serv_auth_ref, serv_token_info = self._validate_token(
-                        request.service_token, request)
-                    request.set_service_headers(serv_auth_ref)
-            except exc.InvalidToken:
-                if self._delay_auth_decision:
-                    self.log.info(
-                        _LI('Invalid service token - deferring reject '
-                            'downstream'))
-                    request.headers['X-Service-Identity-Status'] = 'Invalid'
-                else:
-                    self.log.info(
-                        _LI('Invalid service token - rejecting request'))
-                    self._reject_request()
+        # NOTE(jamielennox): The service status is allowed to be missing if a
+        # service token is not passed. If the service status is missing that's
+        # a valid request. We should find a better way to expose this from the
+        # request object.
+        user_status = request.user_token and request.user_token_valid
+        service_status = request.headers.get('X-Service-Identity-Status',
+                                             'Confirmed')
 
-            request.token_auth = _user_plugin.UserAuthPlugin(user_auth_ref,
-                                                             serv_auth_ref)
+        if not (user_status and service_status == 'Confirmed'):
+            if self._delay_auth_decision:
+                self.log.info(_LI('Deferring reject downstream'))
+            else:
+                self.log.info(_LI('Rejecting request'))
+                self._reject_request()
 
-        except exc.ServiceError as e:
-            self.log.critical(_LC('Unable to obtain admin token: %s'), e)
-            raise webob.exc.HTTPServiceUnavailable()
+        if request.user_token_valid:
+            request.set_user_headers(request.token_auth._user_auth_ref,
+                                     self._include_service_catalog)
+
+        if request.service_token and request.service_token_valid:
+            request.set_service_headers(request.token_auth._serv_auth_ref)
 
         if self.log.isEnabledFor(logging.DEBUG):
-            self.log.debug('Received request from %s' %
+            self.log.debug('Received request from %s',
                            request.token_auth._log_format)
 
     def process_response(self, response):
@@ -589,23 +752,6 @@ class AuthProtocol(_BaseAuthProtocol):
 
         return response
 
-    def _get_user_token_from_request(self, request):
-        """Get token id from request.
-
-        :param env: wsgi request environment
-        :returns: token id
-        :raises exc.InvalidToken: if no token is provided in request
-
-        """
-        token = request.user_token
-
-        if token:
-            return token
-        else:
-            if not self._delay_auth_decision:
-                self.log.debug('Headers: %s', dict(request.headers))
-            raise exc.InvalidToken(_('Unable to find token in headers'))
-
     @property
     def _reject_auth_headers(self):
         header_val = 'Keystone uri=\'%s\'' % self._auth_uri
@@ -657,15 +803,14 @@ class AuthProtocol(_BaseAuthProtocol):
             if cached:
                 return cached
 
-    def _validate_token(self, token, request):
-        """Authenticate user token
+    def _fetch_token(self, token):
+        """Retrieve a token from either a PKI bundle or the identity server.
+
+        :param str token: token id
 
-        :param token: token id
-        :param request: incoming request
-        :returns: uncrypted body of the token if the token is valid
         :raises exc.InvalidToken: if token is rejected
-
         """
+        data = None
         token_hashes = None
 
         try:
@@ -680,59 +825,20 @@ class AuthProtocol(_BaseAuthProtocol):
                     # regardless of initial mechanism used to validate it,
                     # and needs to be checked.
                     self._revocations.check(token_hashes)
-
-                auth_ref = access.AccessInfo.factory(body=data,
-                                                     auth_token=token)
             else:
-                verified = None
+                data = self._validate_offline(token, token_hashes)
+                if not data:
+                    data = self._identity_server.verify_token(token)
 
-                try:
-                    if cms.is_pkiz(token):
-                        verified = self._verify_pkiz_token(token, token_hashes)
-                    elif cms.is_asn1_token(token):
-                        verified = self._verify_signed_token(token,
-                                                             token_hashes)
-                except exceptions.CertificateConfigError:
-                    self.log.warning(_LW('Fetch certificate config failed, '
-                                         'fallback to online validation.'))
-                except exc.RevocationListError:
-                    self.log.warning(_LW('Fetch revocation list failed, '
-                                         'fallback to online validation.'))
-
-                if verified is not None:
-                    data = jsonutils.loads(verified)
-                    auth_ref = access.AccessInfo.factory(body=data,
-                                                         auth_token=token)
-                else:
-                    auth_ref = self._identity_server.verify_token(token)
-                    if auth_ref:
-                        if auth_ref.version == 'v2.0':
-                            data = {'access': auth_ref}
-                        else:  # it's v3.
-                            data = {'token': auth_ref}
-                    else:
-                        data = None
-
-            # 0 seconds of validity means is it valid right now.
-            if auth_ref.will_expire_soon(stale_duration=0):
-                raise exc.InvalidToken(_('Token authorization failed'))
-
-            if _token_is_v2(data) and not auth_ref.project_id:
-                msg = _('Unable to determine service tenancy.')
-                raise exc.InvalidToken(msg)
-
-            self._confirm_token_bind(auth_ref, request)
-
-            if not cached:
                 self._token_cache.store(token_hashes[0], data)
 
-            return auth_ref, data
         except (exceptions.ConnectionRefused, exceptions.RequestTimeout):
             self.log.debug('Token validation failure.', exc_info=True)
             self.log.warning(_LW('Authorization failed for token'))
             raise exc.InvalidToken(_('Token authorization failed'))
-        except exc.ServiceError:
-            raise
+        except exc.ServiceError as e:
+            self.log.critical(_LC('Unable to obtain admin token: %s'), e)
+            raise webob.exc.HTTPServiceUnavailable()
         except Exception:
             self.log.debug('Token validation failure.', exc_info=True)
             if token_hashes:
@@ -740,77 +846,33 @@ class AuthProtocol(_BaseAuthProtocol):
             self.log.warning(_LW('Authorization failed for token'))
             raise exc.InvalidToken(_('Token authorization failed'))
 
-    def _invalid_user_token(self, msg=False):
-        # NOTE(jamielennox): use False as the default so that None is valid
-        if msg is False:
-            msg = _('Token authorization failed')
-
-        raise exc.InvalidToken(msg)
-
-    def _confirm_token_bind(self, auth_ref, req):
-        bind_mode = self._conf_get('enforce_token_bind')
-
-        if bind_mode == _BIND_MODE.DISABLED:
-            return
+        return data
 
+    def _validate_offline(self, token, token_hashes):
         try:
-            if auth_ref.version == 'v2.0':
-                bind = auth_ref['token']['bind']
-            elif auth_ref.version == 'v3':
-                bind = auth_ref['bind']
+            if cms.is_pkiz(token):
+                verified = self._verify_pkiz_token(token, token_hashes)
+            elif cms.is_asn1_token(token):
+                verified = self._verify_signed_token(token, token_hashes)
             else:
-                self._invalid_user_token()
-        except KeyError:
-            bind = {}
-
-        # permissive and strict modes don't require there to be a bind
-        permissive = bind_mode in (_BIND_MODE.PERMISSIVE, _BIND_MODE.STRICT)
-
-        if not bind:
-            if permissive:
-                # no bind provided and none required
+                # Can't do offline validation for this type of token.
                 return
-            else:
-                self.log.info(_LI('No bind information present in token.'))
-                self._invalid_user_token()
-
-        # get the named mode if bind_mode is not one of the predefined
-        if permissive or bind_mode == _BIND_MODE.REQUIRED:
-            name = None
+        except exceptions.CertificateConfigError:
+            self.log.warning(_LW('Fetch certificate config failed, '
+                                 'fallback to online validation.'))
+        except exc.RevocationListError:
+            self.log.warning(_LW('Fetch revocation list failed, '
+                                 'fallback to online validation.'))
         else:
-            name = bind_mode
+            data = jsonutils.loads(verified)
+            return data
 
-        if name and name not in bind:
-            self.log.info(_LI('Named bind mode %s not in bind information'),
-                          name)
-            self._invalid_user_token()
+    def _validate_token(self, auth_ref):
+        super(AuthProtocol, self)._validate_token(auth_ref)
 
-        for bind_type, identifier in six.iteritems(bind):
-            if bind_type == _BIND_MODE.KERBEROS:
-                if req.auth_type != 'negotiate':
-                    self.log.info(_LI('Kerberos credentials required and '
-                                      'not present.'))
-                    self._invalid_user_token()
-
-                if req.remote_user != identifier:
-                    self.log.info(_LI('Kerberos credentials do not match '
-                                      'those in bind.'))
-                    self._invalid_user_token()
-
-                self.log.debug('Kerberos bind authentication successful.')
-
-            elif bind_mode == _BIND_MODE.PERMISSIVE:
-                self.log.debug('Ignoring Unknown bind for permissive mode: '
-                               '%(bind_type)s: %(identifier)s.',
-                               {'bind_type': bind_type,
-                                'identifier': identifier})
-
-            else:
-                self.log.info(
-                    _LI('Couldn`t verify unknown bind: %(bind_type)s: '
-                        '%(identifier)s.'),
-                    {'bind_type': bind_type, 'identifier': identifier})
-                self._invalid_user_token()
+        if auth_ref.version == 'v2.0' and not auth_ref.project_id:
+            msg = _('Unable to determine service tenancy.')
+            raise exc.InvalidToken(msg)
 
     def _cms_verify(self, data, inform=cms.PKI_ASN1_FORM):
         """Verifies the signature of the provided data's IAW CMS syntax.
@@ -900,7 +962,8 @@ class AuthProtocol(_BaseAuthProtocol):
             plugin_kwargs['log'] = self.log
 
         plugin_opts = plugin_class.get_options()
-        CONF.register_opts(plugin_opts, group=group)
+        (self._local_oslo_config or CONF).register_opts(plugin_opts,
+                                                        group=group)
 
         for opt in plugin_opts:
             val = self._conf_get(opt.dest, group=group)
@@ -910,6 +973,45 @@ class AuthProtocol(_BaseAuthProtocol):
 
         return plugin_class.load_from_options(**plugin_kwargs)
 
+    def _determine_project(self):
+        """Determine a project name from all available config sources.
+
+        The sources are checked in the following order:
+
+          1. The paste-deploy config for auth_token middleware
+          2. The keystone_authtoken in the project's config
+          3. The oslo.config CONF.project property
+
+        """
+        try:
+            return self._conf_get('project')
+        except cfg.NoSuchOptError:
+            # Prefer local oslo config object
+            if self._local_oslo_config:
+                return self._local_oslo_config.project
+            try:
+                # CONF.project will exist only if the service uses
+                # oslo.config. It will only be set when the project
+                # calls CONF(...) and when not set oslo.config oddly
+                # raises a NoSuchOptError exception.
+                return CONF.project
+            except cfg.NoSuchOptError:
+                return ''
+
+    def _build_useragent_string(self):
+        project = self._determine_project()
+        if project:
+            project_version = _get_project_version(project)
+            project = '{project}/{project_version} '.format(
+                project=project,
+                project_version=project_version)
+
+        ua_template = ('{project}'
+                       'keystonemiddleware.auth_token/{ksm_version}')
+        return ua_template.format(
+            project=project,
+            ksm_version=_get_project_version('keystonemiddleware'))
+
     def _create_identity_server(self):
         # NOTE(jamielennox): Loading Session here should be exactly the
         # same as calling Session.load_from_conf_options(CONF, GROUP)
@@ -920,7 +1022,8 @@ class AuthProtocol(_BaseAuthProtocol):
             key=self._conf_get('keyfile'),
             cacert=self._conf_get('cafile'),
             insecure=self._conf_get('insecure'),
-            timeout=self._conf_get('http_connect_timeout')
+            timeout=self._conf_get('http_connect_timeout'),
+            user_agent=self._build_useragent_string()
         ))
 
         auth_plugin = self._get_auth_plugin()
@@ -930,6 +1033,7 @@ class AuthProtocol(_BaseAuthProtocol):
             auth=auth_plugin,
             service_type='identity',
             interface='admin',
+            region_name=self._conf_get('region_name'),
             connect_retries=self._conf_get('http_request_max_retries'))
 
         auth_version = self._conf_get('auth_version')
